/************************************************************
 **
 ** pic_op.c : PIC operator generation definitions
 **
 ** Copyright (c) 2004-2005, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <assert.h>
#include <errno.h>
#include <string.h>

#include "../libcore/pf_proc.h"
#include "pic_opfn.h"
#include "pic_stvar.h"
#include "piccolst.h"
#include "pic_inst.h"
#include "pic_op.h"
/*
 * Some assumptions:
 *   On entry to each routine takes a destination and one or two values.
 *   The C promotion rules are done implicitly:
 *     if both operands are the same type, nothing is done
 *     if both are signed, the lesser is promoted to the greater
 *     if the unsigned is greater than the signed, the signed is promoted
 *     if the signed can represent all values of the unsigned, the unsigned
 *       is promoted
 *     failing all else, both are converted to unsigned with the same width
 *       as the signed
 *
 * Notes on different types:
 *   constant   -- 
 *   (s)bit     -- (n) bits masked bit type (n = x & mask)
 *   boolean    -- 0 if set to 0, else 1
 *   integer    -- (n) bytes
 *   (s)integer -- (n) bytes
 *   indirect   -- comes through _ind
 *   lookup     -- comes through lookup table
 */

/*
 * NAME
 *   pic_value_to_w
 *
 * DESCRIPTION
 *   move a value from src to w
 *
 * PARAMETERS
 *   pf  : pfile
 *   src : source value
 *   ofs : offset from start of source value
 *
 * RETURN
 *   none, value is in w
 *
 * NOTES
 *   this is in charge of sign extension and whatnot as necessary
 *     for indirect, dir is used to increment or decrement _fsr
 *     when ofs != 0
 *   _status.z *might* change
 *   _status.c is unchanged
 *   ofs is in bytes
 */
/*
 * these are used to prevent duplicate calls to _multiply or _divide
 * for example:
 *    ch = n % 10
 *    n  = n / 10
 * they are reset when a label or branch is encountered
 * also if any of the values are used as a destination
 */
typedef enum pic_last_value_ {
  PIC_LAST_VALUE_FIRST,
  PIC_LAST_VALUE_DIVISOR = PIC_LAST_VALUE_FIRST,
  PIC_LAST_VALUE_DIVIDEND,
  PIC_LAST_VALUE_MULTIPLIER,
  PIC_LAST_VALUE_MULTIPLICAND,
  PIC_LAST_VALUE_CT
} pic_last_value_t;

static value_t pic_last_values[PIC_LAST_VALUE_CT];

void pic_last_value_set(pic_last_value_t which, value_t n)
{
  assert(which < PIC_LAST_VALUE_CT);

  value_lock(n);
  value_release(pic_last_values[which]);
  pic_last_values[which] = n;
}

void pic_last_values_reset(void)
{
  pic_last_value_t which;
  
  for (which = PIC_LAST_VALUE_FIRST; 
       which < PIC_LAST_VALUE_CT;
       which++) {
    pic_last_value_set(which, VALUE_NONE);
  }
}

static flag_t pic_value_sign_flag_get(value_t val)
{
  return (value_is_signed(val))
    ? VARIABLE_DEF_FLAG_SIGNED
    : VARIABLE_DEF_FLAG_NONE;
}

/* 
 * for a binary op, the result size is the greater of the two
 * operand sizes
 * If the result size is greater than the destination size, the
 * result size is lowered to the destination size
 */
static variable_sz_t pic_result_sz_get(value_t val1, value_t val2, 
    value_t dst)
{
  variable_sz_t sz;

  assert(val1);

  sz = value_byte_sz_get(val1);
  if (val2) {
    variable_sz_t val2_sz;

    val2_sz = value_byte_sz_get(val2);
    if (val2_sz > sz) {
      sz = val2_sz;
    }
  }
  if (dst) {
    variable_sz_t dst_sz;

    dst_sz = value_byte_sz_get(dst);
    if (dst_sz < sz) {
      sz = dst_sz;
    }
  }
  return sz;
}

/*
 * determine the result sign flag based on the promotion rules
 */
static flag_t pic_result_flag_get(value_t val1, value_t val2)
{
  flag_t flag;

  flag = VARIABLE_DEF_FLAG_NONE;
  if (!val2) {
    if (value_is_signed(val1)) {
      flag = VARIABLE_DEF_FLAG_SIGNED;
    }
  } else {
    variable_sz_t val1_sz;
    variable_sz_t val2_sz;

    val1_sz = value_sz_get(val1);
    val2_sz = value_sz_get(val2);
    if (val1_sz < val2_sz) {
      if (value_is_signed(val2)) {
        flag = VARIABLE_DEF_FLAG_SIGNED;
      }
    } else if (val1_sz > val2_sz) {
      if (value_is_signed(val1)) {
        flag = VARIABLE_DEF_FLAG_SIGNED;
      }
    } else if (value_is_signed(val1) && value_is_signed(val2)) {
      flag = VARIABLE_DEF_FLAG_SIGNED;
    }
  }
  return flag;
}

/* bump the indirect pointer */
static void pic_indirect_bump(pfile_t *pf, value_t val, unsigned ofs,
    pic_indirect_change_t change)
{
  if (ofs 
      && (PIC_INDIRECT_CHANGE_NONE != change) 
      && (!val || value_is_indirect(val))
      && (ofs < value_byte_sz_get(val))) {
    pic_instr_append_reg_d(pf, 
        (PIC_INDIRECT_CHANGE_INCR == change)
          ? pic_opcode_incf : pic_opcode_decf,
        "_fsr", pic_opdst_f);
  }
}

static void pic_indirect_bump3(pfile_t *pf, value_t val1, value_t val2,
    value_t dst, unsigned ofs, pic_indirect_change_t change)
{
  if (value_is_indirect(val1)) {
    pic_indirect_bump(pf, val1, ofs, change);
  } else if (value_is_indirect(val2)) {
    pic_indirect_bump(pf, val2, ofs, change);
  } else if (value_is_indirect(dst)) {
    pic_indirect_bump(pf, dst, ofs, change);
  }
}

/* move the value from src into w */
static void pic_op_assign_from_single_bit(pfile_t *pf, value_t dst,
    value_t src);
static void pic_op_assign_from_multi_bit(pfile_t *pf, value_t dst,
    value_t src);
void pic_value_to_w(pfile_t *pf, value_t src, unsigned ofs,
    pic_indirect_change_t change)
{
  if (value_is_lookup(src)) {
    size_t  key;

    key = pic_lookup_is_const(value_variable_get(src), ofs);
    if (-1 == key) {
      label_t    lbl;
      value_t    baseofs;
      pic_code_t code;

      lbl = pic_lookup_label_find(pf, value_variable_get(src), ofs, 
          boolean_false);
      baseofs = value_baseofs_get(src);
      /* need to find a way to set PCLATH to the high byte of the dst */
      code = pic_instr_append(pf, pic_opcode_movlw);
      pic_code_brdst_set(code, lbl);
      pic_code_ofs_set(code, 1);
      pic_instr_append_reg(pf, pic_opcode_movwf, "_pclath");
      pic_instr_append_f_d(pf, pic_opcode_movf, baseofs, 0, pic_opdst_w);
      pic_instr_append_n(pf, pic_opcode_call, lbl);
      label_release(lbl);
    } else {
      pic_instr_append_w_kn(pf, pic_opcode_movlw, key);
    }
  } else if (value_is_const(src)) {
    pic_instr_append_w_kn(pf, pic_opcode_movlw,
        value_const_get(src) >> (8 * ofs));
  } else if (value_is_single_bit(src)) {
    pic_op_assign_from_single_bit(pf, VALUE_NONE, src);
  } else if (value_is_multi_bit(src)) {
    pic_op_assign_from_multi_bit(pf, VALUE_NONE, src);
  } else {
    pic_indirect_bump(pf, src, ofs, change);
    pic_instr_append_f_d(pf, pic_opcode_movf, src, ofs, pic_opdst_w);
  }
}

/* put w into dst */
static void pic_w_to_value(pfile_t *pf, value_t dst, unsigned ofs,
    pic_indirect_change_t change)
{
  if (value_is_lookup(dst)) {
    assert(0);
  } else if (value_is_const(dst)) {
    assert(0);
  } else {
    assert(!value_is_bit(dst));
    if (ofs) {
      pic_indirect_bump(pf, dst, ofs, change);
    }
    pic_instr_append_f(pf, pic_opcode_movwf, dst, ofs);
  }
}

/*
 * set all of dst to w (dst[0] = w, dst[1..n] = 0)
 *
 * nb: if dst is indirect, it must be setup *before* getting here! 
 */
void pic_set_to_w(pfile_t *pf, value_t dst)
{
  size_t ii;

  if (value_is_bit(dst)) {
    value_t tmp;

    tmp = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
    pic_set_to_w(pf, tmp);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else if (dst) {
    pic_instr_append_f(pf, pic_opcode_movwf, dst, 0);
    for (ii = 1; ii < value_sz_get(dst); ii++) {
      pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
      pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
    }
  } else {
    /* just set the Z flag */
    pic_instr_append_w_kn(pf, pic_opcode_andlw, 0xff);
  }
}

/*
 * NAME
 *   pic_indirect_setup
 *
 * DESCRIPTION
 *   setup irp:fsr
 *
 * PARAMETERS
 *   pf    : file
 *   val   : value
 *   offset: offset from start of value
 *
 * RETURN
 *
 * NOTES
 *   1. A value is assumed to not span banks
 *   2. Normally this is called if val.baseofs is non-const
 */
void pic_indirect_setup(pfile_t *pf, value_t val, size_t offset)
{
  value_t baseofs;
  value_t ind;

  ind = pfile_value_find(pf, pfile_log_err, "_ind");
  if (ind) {
    value_assign_ct_bump(ind, ctr_bump_incr);
    value_use_ct_bump(ind, ctr_bump_incr);
    value_release(ind);
  }

  baseofs = value_baseofs_get(val);
  if (value_is_bit(val)) {
    offset += (value_bit_offset_get(val) + 7) / 8;
  }

  if (value_is_indirect(val)) {
    /* STATUS<IRP>:FSR = *value */
    pic_opcode_t op;
    value_t      tmp;

    offset += value_const_get(baseofs);
    tmp = value_clone(val);
    value_baseofs_set(tmp, VALUE_NONE);

    pic_instr_append_f(pf, pic_opcode_movlw, tmp, offset);
    pic_instr_append_reg(pf, pic_opcode_movwf, "_fsr");
    op = (value_base_get(tmp) >= 256)
      ? pic_opcode_irp_set
      : pic_opcode_irp_clr;
    pic_instr_append(pf, op);
    /* finally, add any additional offset to _fsr */
    if (!value_is_const(baseofs)) {
      pic_instr_append_f_d(pf, pic_opcode_movf, baseofs, 0, pic_opdst_w);
      pic_instr_append_reg_d(pf, pic_opcode_addwf, "_fsr", pic_opdst_f);
    }
    value_release(tmp);
  } else {
    variable_base_t base;

    pic_instr_append(pf, 
       (variable_base_get(value_variable_get(val), 0) >= 256) 
       ? pic_opcode_irp_set : pic_opcode_irp_clr);
    base = value_base_get(val) + offset;
    if (value_is_const(baseofs)) {
      /* fsr = value_base_get(value) + value_const_get(baseofs) + sz */
      base += value_const_get(baseofs);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, base);
      pic_instr_append_reg(pf, pic_opcode_movwf, "_fsr");
    } else {
      /* fsr = base (+ sz - 1) */
      pic_instr_append_w_kn(pf, pic_opcode_movlw, base);
      pic_instr_append_reg_d(pf, pic_opcode_movwf, "_fsr", 0);
      if (VALUE_NONE != baseofs) {
        /* w = baseofs */
        pic_instr_append_f_d(pf, pic_opcode_movf, baseofs, 0, pic_opdst_w);
        /* _fsr += w */
        pic_instr_append_reg_d(pf, pic_opcode_addwf, "_fsr", pic_opdst_f);
      }
    }
  }
}

/* setup irp:fsr for either val1 or val2 or neither */
static void pic_indirect_setup3(pfile_t *pf, value_t dst, 
    value_t val1, value_t val2)
{
  assert(value_is_same(dst, val1)
      || !(value_is_indirect(dst) && value_is_indirect(val1)));
  assert(value_is_same(dst, val2)
      || !(value_is_indirect(dst) && value_is_indirect(val2)));
  assert(value_is_same(val1, val2)
      || !(value_is_indirect(val1) && value_is_indirect(val2)));

  if (value_is_indirect(dst)) {
    pic_indirect_setup(pf, dst, 0);
  } else if (value_is_indirect(val1)) {
    pic_indirect_setup(pf, val1, 0);
  } else if (value_is_indirect(val2)) {
    pic_indirect_setup(pf, val2, 0);
  }
}

/* determine if the result of a binary operation is signed */
static boolean_t pic_result_is_signed(value_t val1, value_t val2)
{
  return (VARIABLE_DEF_FLAG_SIGNED == pic_result_flag_get(val1, val2));
}

/* given two values, determine the correct variable definition */
static variable_def_t pic_result_vdef_get(value_t val1, value_t val2, 
    value_t dst)
{
  variable_sz_t sz;
  flag_t        flag;

  sz   = pic_result_sz_get(val1, val2, dst);
  flag = pic_result_flag_get(val1, val2);

  return variable_def_alloc(0, variable_def_type_integer, flag, sz);
}

/*
 * perform sign extension
 * if dst is indirect, it must be pointing to the last set byte
 */
static void pic_sign_extend(pfile_t *pf, boolean_t is_signed, value_t dst, 
    size_t ofs)
{
  size_t dst_sz;

  dst_sz = value_sz_get(dst);
  if (ofs < dst_sz) {
    /* sign extend */
    pic_opcode_t pop;

    if (is_signed) {
      pic_instr_append(pf, pic_opcode_clrw);
      pic_instr_append_f_bn(pf, pic_opcode_btfsc, dst, ofs - 1, 7);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 255);
      pop = pic_opcode_movwf;
    } else {
      pop = pic_opcode_clrf;
    }
    while (ofs < dst_sz) {
      pic_indirect_bump(pf, dst, ofs, PIC_INDIRECT_CHANGE_INCR);
      pic_instr_append_f(pf, pop, dst, ofs);
      ofs++;
    }
  }
}

/* 
 * allocate a pic temp variable & assign val to it. The size of the result
 * is less{dst, greater{val1, val2}}
 */
static value_t pic_var_temp_get_assign(pfile_t *pf, value_t val, value_t val2,
    value_t dst)
{
  value_t        tmp;
  size_t         sz;

  sz = pic_result_sz_get(val, val2, dst);
  tmp = pic_var_temp_get(pf, pic_result_flag_get(val, val2), sz);
  pic_op(pf, operator_assign, tmp, val, VALUE_NONE);
  return tmp;
}

/* return a 1-byte value that holds the sign of val */
static value_t pic_sign_get(pfile_t *pf, value_t val, size_t ofs)
{
  value_t sign;

  sign = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
  if (value_is_signed(val)) {
    pic_instr_append(pf, pic_opcode_clrw);
    pic_instr_append_f_bn(pf, pic_opcode_btfsc, val, ofs, 7);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 255);
    pic_instr_append_f(pf, pic_opcode_movwf, sign, 0);
  } else {
    pic_instr_append_f(pf, pic_opcode_clrf, sign, 0);
  }
  return sign;
}

/*
 * this is used enough to warrent its own function. often it's
 * assigning _status:_z to a bit
 * invert is TRUE to invert the source bit 
 * src    is VALUE_NONE for _status:_z
 */
typedef enum {
  PIC_ASSIGN_SINGLE_BIT_ACTION_NONE,
  PIC_ASSIGN_SINGLE_BIT_ACTION_INVERT
} pic_assign_single_bit_action_t;

static void pic_op_assign_single_bit_to_single_bit(pfile_t *pf,
    value_t dst, value_t src, pic_assign_single_bit_action_t action)
{
  pic_opcode_t pop;

  pop = pic_opcode_btfss;
  switch (action) {
    case PIC_ASSIGN_SINGLE_BIT_ACTION_NONE:
      break;
    case PIC_ASSIGN_SINGLE_BIT_ACTION_INVERT:
      pop = pic_opcode_btfsc;
      break;
  }

  if (value_is_volatile(dst)
      || value_is_same(dst, src)) {
    label_t lbl_set;
    label_t lbl_skip;

    lbl_set  = pfile_label_alloc(pf, 0);
    lbl_skip = pfile_label_alloc(pf, 0);

    if (src) {
      pic_instr_append_f(pf, pop, src, 0);
    } else {
      pic_instr_append_reg_flag(pf, 
          (pic_opcode_btfsc == pop) ? pic_opcode_btfss : pic_opcode_btfsc, 
          "_status", "_z");
    }
    pic_instr_append_n(pf, pic_opcode_goto, lbl_set);
    pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
    pic_instr_append_n(pf, pic_opcode_goto, lbl_skip);
    pic_instr_append_label(pf, lbl_set);
    pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
    pic_instr_append_label(pf, lbl_skip);
    label_release(lbl_skip);
    label_release(lbl_set);
  } else {
    pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
    if (src) {
      pic_instr_append_f(pf, pop, src, 0);
    } else {
      pic_instr_append_reg_flag(pf, 
          (pic_opcode_btfsc == pop) ? pic_opcode_btfss : pic_opcode_btfsc, 
          "_status", "_z");
    }
    pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
  }
}

/* nb: this works for both boolean *and* straight bit values */
static void pic_op_assign_to_single_bit(pfile_t *pf, value_t dst, 
    value_t src)
{
  /*
   * first the trivial case : src is constant, simply set or clear the
   * appropriate bit in dst
   */
  if (value_is_const(src)) {
    /* this is much simplified! */
    pic_opcode_t pop;

    if (variable_def_type_boolean == value_type_get(dst)) {
      pop = (value_const_get(src)) ? pic_opcode_bsf : pic_opcode_bcf;
    } else {
      pop = (value_const_get(src) & 0x01) ? pic_opcode_bsf : pic_opcode_bcf;
    }
    if (value_is_indirect(dst)) {
      pic_indirect_setup(pf, dst, value_bit_offset_get(dst) / 8);
    }
    pic_instr_append_f_bn(pf, pop, dst, 0, 0);
  /*
   * second, the easy case : src is a single bit, simply transfer the bit
   * from src to dst
   */
  } else if (value_is_single_bit(src)) {
    /* this is also simplified */
    pic_indirect_setup3(pf, src, dst, VALUE_NONE);
    pic_op_assign_single_bit_to_single_bit(pf, dst, src, 
        PIC_ASSIGN_SINGLE_BIT_ACTION_NONE);
  /*
   * finally, the complex case; src is an unknown entity. first
   * assign it to a temporary place to make it a simply variable,
   * then for boolean test if it's 0, for other test if the low bit
   * is 0
   */
  } else {
    value_t   t_src;

    if (value_is_boolean(dst) && !value_is_indirect(dst)) {
      pic_op(pf, operator_logical, VALUE_NONE, src, VALUE_NONE);
      t_src = VALUE_NONE;
    } else {
      t_src   = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
      pic_op(pf, operator_assign, t_src, src, VALUE_NONE);
    }

    pic_indirect_setup3(pf, t_src, dst, VALUE_NONE);
    pic_op_assign_single_bit_to_single_bit(pf, dst, t_src, 
        PIC_ASSIGN_SINGLE_BIT_ACTION_NONE);
    if (t_src) {
      pic_var_temp_release(pf, t_src);
    }
  }
}


/*
 * dst &= dst_mask
 * src &= src_mask
 * src <<= dst_ofs
 * dst  |= src
 */
static void pic_op_assign_to_multi_bit(pfile_t *pf, value_t dst, 
    value_t src)
{
  if (value_is_bit(src)) {
    /*
     * src --> tmp
     * tmp --> dst
     */
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, src, VALUE_NONE, VALUE_NONE);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else {
    size_t    dst_sz;   /* size (in bits)       */
    size_t    dst_ofs;  /* offset (in bits)     */
    size_t    dst_mask; /* mask for destination */
    size_t    src_mask; /* mask for source      */
    boolean_t dst_is_volatile;
    value_t   d_mask;
    value_t   v_dst;

    dst_sz          = value_sz_get(dst);
    dst_ofs         = value_bit_offset_get(dst);
    src_mask        = (1 << dst_sz) - 1;
    dst_mask        = ~(src_mask << dst_ofs);
    dst_is_volatile = value_is_volatile(dst);
    dst             = value_alloc(variable_master_get(value_variable_get(dst)));
    d_mask          = pfile_constant_get(pf, dst_mask, value_def_get(dst));

    if (dst_is_volatile) {
      v_dst = pic_var_temp_get_def(pf, value_def_get(dst));
      pic_op(pf, operator_andb, v_dst, dst, d_mask);
    } else {
      v_dst = dst;
      pic_op(pf, operator_andb, v_dst, d_mask, VALUE_NONE);
    }

    if (value_is_const(src)) {
      variable_const_t n;
      value_t          n_val;

      n = (value_const_get(src) & src_mask) << dst_ofs;
      n_val = pfile_constant_get(pf, n, value_def_get(dst));
      pic_op(pf, operator_orb, v_dst, n_val, VALUE_NONE);
      value_release(n_val);
    } else {
      value_t   t_src;
      value_t   s_mask;
      value_t   s_shift;

      /* dst.master = (dst.master & dst_mask) | ((src & src_mask) << bitofs) */
      s_shift         = pfile_constant_get(pf, dst_ofs, value_def_get(dst));
      s_mask          = pfile_constant_get(pf, src_mask, value_def_get(dst));
      t_src           = pic_var_temp_get_def(pf, value_def_get(dst));

      pic_op(pf, operator_andb, t_src, src, s_mask);
      pic_op(pf, operator_shift_left, t_src, src, s_shift);
      pic_op(pf, operator_orb,  dst, v_dst, t_src);
      pic_var_temp_release(pf, t_src);
      value_release(s_mask);
      value_release(s_shift);
    }
    if (v_dst != dst) {
      pic_var_temp_release(pf, v_dst);
    }
    value_release(d_mask);
    value_release(dst);
  }
}

static void pic_op_assign_from_multi_bit(pfile_t *pf, value_t dst,
    value_t src)
{
  size_t  src_ofs;
  size_t  src_sz;
  value_t v_mask;
  value_t t_src;
  value_t t_dst;

  src_ofs  = value_bit_offset_get(src);
  src_sz   = value_sz_get(src);

  src      = value_alloc(variable_master_get(value_variable_get(src)));
  if (src_ofs) {
    value_t v_shift;

    t_src    = pic_var_temp_get_def(pf, value_def_get(src));
    v_shift  = pfile_constant_get(pf, src_ofs, value_def_get(src));
    pic_op(pf, operator_shift_right, t_src, src,  v_shift);
    value_release(v_shift);
  } else {
    t_src = src;
  }
  v_mask = pfile_constant_get(pf, (1 << src_sz) - 1, value_def_get(src));

  if (!dst) {
    variable_def_t def;

    def = variable_def_alloc(0, variable_def_type_integer, 0, 1);
    t_dst = pic_var_temp_get_def(pf, def);
  }

  pic_op(pf, operator_andb, t_dst, t_src, v_mask);
  if (dst != t_dst) {
    pic_op(pf, operator_assign, VALUE_NONE, t_dst, VALUE_NONE);
    pic_var_temp_release(pf, t_dst);
  }
  if (t_src != src) {
    pic_var_temp_release(pf, t_src);
  }
  value_release(v_mask);
  value_release(src);
}

static void pic_op_assign_from_single_bit(pfile_t *pf, value_t dst,
    value_t src)
{
  size_t src_byte_ofs;
  size_t src_bit_ofs;

  src_bit_ofs  = value_bit_offset_get(src);
  src_byte_ofs = src_bit_ofs / 8;
  src_bit_ofs &= 7;

  src = value_alloc(variable_master_get(value_variable_get(src)));
  if (value_is_indirect(src)) {
    pic_indirect_setup(pf, src, src_byte_ofs);
  }
  pic_instr_append(pf, pic_opcode_clrw);
  pic_instr_append_f_bn(pf, pic_opcode_btfsc, src, src_byte_ofs, 
      src_bit_ofs);
  pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
  if (dst) {
    pic_set_to_w(pf, dst);
  }
  value_release(src);
}

static void pic_op_assign_from_lookup(pfile_t *pf, value_t dst,
    value_t src)
{
  size_t  dst_sz;
  size_t  src_sz;
  size_t  sz;
  size_t  ii;
  size_t  key_pv;

  dst_sz = value_sz_get(dst);
  src_sz = value_sz_get(src);
  if (dst_sz < src_sz) {
    sz = dst_sz;
  } else {
    sz = src_sz;
  }
  if (value_is_indirect(dst)) {
    pic_indirect_setup(pf, dst, 0);
  }
  for (ii = 0, key_pv = -1; ii < sz; ii++) {
    size_t key;

    key = pic_lookup_is_const(value_variable_get(src), ii);
    if (-1 == key) {
      pic_value_to_w(pf, src, ii, PIC_INDIRECT_CHANGE_NONE);
      pic_w_to_value(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
    } else if (0 != key) {
      if (key != key_pv) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, key);
      }
      pic_w_to_value(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
    } else {
      pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
      pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
    }
    key_pv = key;
  }
  pic_sign_extend(pf, value_is_signed(src), dst, ii);
}

/* nb: this needs to be special cased because the internal format
 *     for a const may not be the same as the internal format of
 *     an array (for the PIC, it's not. const is kept big endian,
 *     and the PIC stores data little endian)
 */
static void pic_op_assign_from_const_array(pfile_t *pf, value_t dst,
    value_t src)
{
  variable_t            src_var;
  variable_def_t        src_def;
  variable_def_member_t src_mbr;
  size_t                dst_sz;
  size_t                dst_ofs;
  size_t                src_ct; /* # of elements in the source      */
  size_t                src_sz; /* per *element* size of the source */
  size_t                key_pv; /* last key out */
  size_t                ii;

  src_mbr = variable_def_member_get(value_def_get(src));
  src_def = variable_def_member_def_get(src_mbr);
  dst_sz  = value_sz_get(dst);
  src_ct  = variable_def_member_ct_get(src_mbr);
  src_sz  = variable_def_sz_get(src_def);
  src_var = value_variable_get(src);
  
  assert(value_sz_get(src) == dst_sz);
  assert(!value_baseofs_get(src));

  if (value_is_indirect(dst)) {
    pic_indirect_setup(pf, dst, 0);
  }

  dst_ofs = 0;
  for (ii = 0, key_pv = -1; (ii < src_ct) && (dst_ofs < dst_sz); ii++) {
    size_t           jj;
    variable_const_t n;

    n = variable_const_get(src_var, src_def, ii * src_sz);

    for (jj = 0; (jj < src_sz) && (dst_ofs < dst_sz); jj++, dst_ofs++) {
      size_t key;

      key = n & 0xff;
      n >>= 8;
      pic_indirect_bump(pf, dst, dst_ofs, PIC_INDIRECT_CHANGE_INCR);
      if (0 == key) {
        pic_instr_append_f(pf, pic_opcode_clrf, dst, dst_ofs);
      } else {
        if (key != key_pv) {
          pic_instr_append_w_kn(pf, pic_opcode_movlw, key);
        }
        pic_instr_append_f(pf, pic_opcode_movwf, dst, dst_ofs);
        key_pv = key;
      }
    }
  }
}

static void pic_op_assign_from_const(pfile_t *pf, value_t dst,
    value_t src)
{
  if (value_is_array(src)) {
    pic_op_assign_from_const_array(pf, dst, src);
  } else {
    size_t           dst_sz;
    size_t           src_sz;
    size_t           sz;
    size_t           ii;
    variable_const_t n;
    size_t           key_pv;

    dst_sz = value_sz_get(dst);
    src_sz = value_sz_get(src);
    sz = (dst_sz < src_sz) ? dst_sz : src_sz;
    if (value_is_indirect(dst)) {
      pic_indirect_setup(pf, dst, 0);
    }
    n = value_const_get(src);
    for (ii = 0, key_pv = -1; ii < sz; ii++) {
      size_t key;

      if (ii) {
        n >>= 8;
      }
      key = n & 0xff;
      if (0 == key) {
        pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
      } else {
        if (key != key_pv) {
          pic_value_to_w(pf, src, ii, PIC_INDIRECT_CHANGE_NONE);
        }
        pic_w_to_value(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
      }
      key_pv = key;
    }
    if (ii < dst_sz) {
      if (value_is_signed(src) && (key_pv & 0x80)) {
        if (255 != key_pv) {
          pic_instr_append_w_kn(pf, pic_opcode_movlw, 255);
        }
        while (ii < dst_sz) {
          pic_w_to_value(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
          ii++;
        }
      } else {
        while (ii < dst_sz) {
          pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
          pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
          ii++;
        }
      }
    }
  }
}

static void pic_op_assign_from_function(pfile_t *pf, value_t dst,
  value_t src)
{
  variable_def_t src_def;
  variable_def_t dst_def;
  pic_code_t     code;
  pfile_proc_t  *proc;
  label_t        lbl;

  assert(value_is_pointer(dst));
  src_def = value_def_get(src);
  dst_def = value_def_get(dst);
  /* dereference dst def */
  dst_def = variable_def_member_def_get(variable_def_member_get(dst_def));
  assert(variable_def_is_same(src_def, dst_def));

  proc = value_proc_get(src);
  lbl  = pfile_proc_label_get(proc);

  if (value_is_indirect(dst)) {
    pic_indirect_setup(pf, dst, 0);
  }
  code = pic_instr_append(pf, pic_opcode_movlw);
  pic_code_brdst_set(code, lbl);
  pic_instr_append_f(pf, pic_opcode_movwf, dst, 0);
  pic_indirect_bump(pf, dst, 1, PIC_INDIRECT_CHANGE_INCR);
  code = pic_instr_append(pf, pic_opcode_movlw);
  pic_code_brdst_set(code, lbl);
  pic_code_ofs_set(code, 1);
  pic_instr_append_f(pf, pic_opcode_movwf, dst, 1);
}

static void pic_op_assign(pfile_t *pf, operator_t op, value_t dst, 
    value_t src, value_t dummy)
{
  if (!value_variable_get(dst)) {
    /* this is a special case -- return value in W */
    /* assert(value_is_const(src) || (1 == value_byte_sz_get(src))); */
    if (value_is_indirect(src)) {
      pic_indirect_setup(pf, src, 0);
    }
    pic_value_to_w(pf, src, 0, PIC_INDIRECT_CHANGE_INCR);
  } else if (value_is_indirect(dst) && value_is_indirect(src)) {
    /* 
     * src --> tmp
     * tmp --> dst
     */ 
    value_t tmp;

    tmp = pic_var_temp_get_def(pf, value_def_get(src));
    pic_op(pf, operator_assign, tmp, src, VALUE_NONE);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else {
    if (value_is_single_bit(dst)) {
      /* assign anything to a single bit */
      pic_op_assign_to_single_bit(pf, dst, src);
    } else if (value_is_bit(dst)) {
      /* assign anything to a multi bit  */
      pic_op_assign_to_multi_bit(pf, dst, src);
    } else if (value_is_const(src)) {
      /* assign a const to any non-bit */
      pic_op_assign_from_const(pf, dst, src);
    } else if (value_is_single_bit(src)) {
      /* assign a single bit to any non-bit */
      pic_op_assign_from_single_bit(pf, dst, src);
    } else if (value_is_bit(src)) {
      /* assign a multi bit to any non-bit */
      pic_op_assign_from_multi_bit(pf, dst, src);
    } else if (value_is_lookup(src)) {
      /* assign a lookup to any non-bit */
      pic_op_assign_from_lookup(pf, dst, src);
    } else if (value_is_function(src)) {
      pic_op_assign_from_function(pf, dst, src);
    } else {
      /* handle the remainder */
      size_t src_sz;
      size_t dst_sz;
      size_t sz;
      size_t ii;

      src_sz = value_sz_get(src);
      dst_sz = value_sz_get(dst);
      sz     = (src_sz < dst_sz) ? src_sz : dst_sz;
      pic_indirect_setup3(pf, src, dst, VALUE_NONE);
      for (ii = 0; ii < sz; ii++) {
        pic_value_to_w(pf, src, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_w_to_value(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
      }
      pic_sign_extend(pf, value_is_signed(src), dst, ii);
    }
  }
}

/*
 * operator_logical = !!a (result is 0 if a is 0, 1 if a is anything else)
 * operator_not     =  !a (result is 1 if a is 0, 0 if a is anything else)
 * nb: if dst is VALUE_NONE, the result is left in W; this generally means
 *     someone only wanted the _status:_z set
 */
static void pic_op_logical_not(pfile_t *pf, operator_t op, value_t dst, 
    value_t val, value_t dummy)
{
  assert(!value_is_const(val)); /* this would be handled elsewhere! */
  if (value_is_single_bit(val)) {
    pic_opcode_t pop;

    pop = (operator_logical == op)
      ? pic_opcode_btfsc
      : pic_opcode_btfss;
    if (!dst) {
      /* set Z as necessary */
      pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_z");
      pic_instr_append_f(pf, pop, val, 0);
      pic_instr_append_reg_flag(pf, pic_opcode_bsf, "_status", "_z");
    } else {
      if (value_is_indirect(dst)) {
        pic_indirect_setup(pf, dst, 0);
      }
      if (value_is_single_bit(dst)) {
        pic_op_assign_single_bit_to_single_bit(pf,
            dst, val, 
            (operator_notl == op)
            ? PIC_ASSIGN_SINGLE_BIT_ACTION_INVERT
            : PIC_ASSIGN_SINGLE_BIT_ACTION_NONE);
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
        pic_instr_append_f(pf, pop, val, 0);
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
        pic_set_to_w(pf, dst);
      }
    }
  } else if (value_is_bit(val) || value_is_lookup(val)) {
    /* assign this to a temporary value & work on it */
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val, VALUE_NONE, dst);
    pic_op(pf, op, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else {
    size_t sz;
    size_t ii;

    if (value_is_indirect(val)) {
      pic_indirect_setup(pf, val, 0);
    }
    sz = value_sz_get(val);
    for (ii = 0; ii < sz; ii++) {
      if (0 == ii) {
        pic_instr_append_f_d(pf, pic_opcode_movf, val, ii, pic_opdst_w);
      } else {
        pic_indirect_bump(pf, val, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_f_d(pf, pic_opcode_iorwf, val, ii, pic_opdst_w);
      }
    }
    if (operator_notl == op) {
      /* if w is zero, change to 1, otherwise keep it 0 */
      /* first, invert _status:_z */
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
    } else if (operator_logical == op) {
      /* 
       * if dst exists & w is non-zero, change it to 1; otherwise
       * the caller is only interested in _status:_z and that's
       * already correct
       */
      if (dst) {
        pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      }
    }
    if (dst) {
      pic_set_to_w(pf, dst);
    }
  }
}

/* 2s complement negation:
 *   x = ~x + 1
 * if x is a single bit, this is a no-op  (but that's detected elsewhere)
 */
static void pic_op_neg(pfile_t *pf, operator_t op, value_t dst, value_t val,
    value_t dummy)
{
  assert(!value_is_const(val));
  if (value_is_volatile(dst) 
      || value_is_bit(dst)) {
    value_t tmp;
    size_t  sz;
    /*
     * there's no reason to do more work than necessary, so only
     * work on the smaller length of (val, dst)
     */
    sz = pic_result_sz_get(val, VALUE_NONE, dst);
    tmp = pic_var_temp_get(pf, pic_value_sign_flag_get(val), sz);
    pic_op(pf, operator_cmpb, tmp, val, VALUE_NONE);
    pic_op(pf, operator_incr, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else {
    value_t work_val;

    work_val = value_clone(dst);
    value_def_set(work_val, pic_result_vdef_get(val, VALUE_NONE, dst));
    pic_op(pf, operator_cmpb, work_val, val, VALUE_NONE);
    pic_op(pf, operator_incr, work_val, work_val, VALUE_NONE);
    pic_sign_extend(pf, value_is_signed(val), dst,
        value_byte_sz_get(work_val));
    value_release(work_val);
  }
}

/* 1s complement -- if val's a constant this should be an assignment! */
static void pic_op_cmpb(pfile_t *pf, operator_t op, value_t dst, value_t val,
    value_t dummy)
{
  assert(!value_is_const(val));
  if (value_is_single_bit(val)) {
    /* this is the same as a logical not */
    pic_op(pf, operator_notl, dst, val, dummy);
  } else if (value_is_bit(dst)) {
    value_t tmp;

    tmp = pic_var_temp_get_def(pf, value_def_get(dst));
    pic_op(pf, op, tmp, val, dummy);
    pic_op(pf, operator_assign, dst, tmp, dummy);
    pic_var_temp_release(pf, tmp);
  } else if (value_is_bit(val) 
      || value_is_lookup(val)
      || (!value_is_same(dst, val)
        && value_is_indirect(dst) && value_is_indirect(dst))) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val, VALUE_NONE, dst);
    pic_op(pf, op, dst, tmp, dummy);
    pic_var_temp_release(pf, tmp);
  } else {
    size_t      ii;
    size_t      sz;
    pic_opdst_t op_dst;

    sz = pic_result_sz_get(val, VALUE_NONE, dst);
    pic_indirect_setup3(pf, dst, val, VALUE_NONE);
    op_dst = value_is_same(dst, val) ? pic_opdst_f : pic_opdst_w;
    for (ii = 0; ii < sz; ii++) {
      pic_indirect_bump3(pf, val, VALUE_NONE, dst, ii,
          PIC_INDIRECT_CHANGE_INCR);
      pic_instr_append_f_d(pf, pic_opcode_comf, val, ii, op_dst);
      if (pic_opdst_w == op_dst) {
        pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
      }
    }
    pic_sign_extend(pf, value_is_signed(val), dst, ii);
  }
}

static void pic_op_add_sub(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2)
{
  variable_sz_t sz;

  assert(!(value_is_const(val1) && value_is_const(val2)));
  sz = pic_result_sz_get(val1, val2, dst);
  if ((operator_add == op)
      && value_is_const(val1) 
      && ((1 == value_const_get(val1)) || (-1 == value_const_get(val1)))) {
    /* x = 1 + y --> x = incr(y)                */
    /* x = -1 + y --> x = y - 1 --> x = decr(y) */
    pic_op(pf, (1 == value_const_get(val1)) ? operator_incr : operator_decr, 
          dst, val2, VALUE_NONE);
  } else if (value_is_const(val2) && 
      ((1 == value_const_get(val2)) || (-1 == value_const_get(val2)))) {
    /* x = y [+/-] 1 --> x = y[++/--] */
    operator_t op2;

    if (operator_add == op) {
      op2 = (1 == value_const_get(val2)) ? operator_incr : operator_decr;
    } else {
      op2 = (1 == value_const_get(val2)) ? operator_decr : operator_incr;
    }
    pic_op(pf, op2, dst, val1, VALUE_NONE);
  } else if (value_is_same(val1, val2)) {
    /* x = y + y --> x = y << 1 */
    /* x = y - y *must be* optimized out elsewhere! */
    value_t c_one;

    assert(operator_add == op);
    c_one = pfile_constant_get(pf, 1, VARIABLE_DEF_NONE);
    pic_op(pf, op, dst, val1, c_one);
    value_release(c_one);
  } else if (value_is_bit(dst)
      || value_is_volatile(dst)
      || !value_is_same(dst, val1)
      || value_is_bit(val1)
      || value_is_lookup(val1)) {
    /* 
     * work must be done in a temporary.
     */
    value_t work_val;

    work_val = pic_var_temp_get_def(pf, 
        pic_result_vdef_get(val1, val2, dst));
    pic_op(pf, operator_assign, work_val, val1, VALUE_NONE);
    pic_op(pf, op, work_val, work_val, val2);
    pic_op(pf, operator_assign, dst, work_val, VALUE_NONE);
    pic_var_temp_release(pf, work_val);
  } else if (value_is_bit(val2) 
      || value_is_lookup(val2)
      || (value_is_indirect(val2)
        && (value_is_indirect(dst) && !value_is_same(val2, dst)))) {
    /* tmp = val2; dst = val1 +/- tmp */
    value_t work_val;

    work_val = pic_var_temp_get_def(pf, 
        pic_result_vdef_get(val1, val2, dst));
    pic_op(pf, operator_assign, work_val, val2, VALUE_NONE);
    pic_op(pf, op, work_val, val1, work_val);
    pic_op(pf, operator_assign, dst, work_val, VALUE_NONE);
    pic_var_temp_release(pf, work_val);
  } else if (value_is_const(val2)) {
    /* val1 must always hold the constant! */
    if (operator_sub == op) {
      val2 = pfile_constant_get(pf, -value_const_get(val2), 
          value_def_get(val2));
      op = operator_add;
    } else {
      value_lock(val2);
    }
    pic_op(pf, op, dst, val2, val1);
    value_release(val2);
  } else {
    /* at this point we've x += y or x -= y */
    size_t       ii;
    pic_opdst_t  op_dst;
    value_t      sign;         /* sign (for sign extension) */
    pic_opcode_t pop_action_f; /* add or sub     */
    pic_opcode_t pop_skip;     /* btfss or btfsc */

    sign = VALUE_NONE;

    if (operator_add == op) {
      /* the operation is far shorter if addition has the constant in val1 */
      pop_action_f = pic_opcode_addwf;
      pop_skip     = pic_opcode_btfsc;
      if ((value_is_const(val2) || value_is_same(dst, val2))) {
        value_t tmp;

        tmp  = val1;
        val1 = val2;
        val2 = tmp;
      }
    } else {
      pop_action_f = pic_opcode_subwf;
      pop_skip     = pic_opcode_btfss;
    }
    pic_indirect_setup3(pf, dst, val1, val2);
    op_dst = (value_is_same(dst, val1)) ? pic_opdst_f : pic_opdst_w;
    for (ii = 0; ii < sz; ii++) {
      /* setup signs if necessary */
      if (ii == value_sz_get(val1)) {
        assert(!sign);
        sign = pic_sign_get(pf, val1, ii - 1);
      } else if (ii == value_sz_get(val2)) {
        assert(!sign);
        sign = pic_sign_get(pf, val2, ii - 1);
      }
      /* val2[x] --> w (don't change _status:_c!) */
      if (ii < value_sz_get(val2)) {
        pic_indirect_bump(pf, val2, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_f_d(pf, pic_opcode_movf, val2, ii, pic_opdst_w);
      } else {
        assert(sign);
        pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
      }
      if (ii) {
        boolean_t    bank_diff;
        pic_opcode_t inc_op;

        pic_instr_append_reg_flag(pf, pop_skip, "_status", "_c");
        /* 
         * determine if the banks are different, if so modify the 
         * behaviour below 
         */
        bank_diff = boolean_false;
        if (ii < value_sz_get(val1)) {
          if (value_is_indirect(val1)) {
            bank_diff = boolean_false;
          } else if (ii < value_sz_get(val2)) {
            bank_diff = !value_is_indirect(val2)
              && pic_are_variable_databits_different(val1, val2);
          } else if (sign) {
            bank_diff = pic_are_variable_databits_different(val1, sign);
          } else {
            bank_diff = boolean_false;
          }
        } else if (ii < value_sz_get(val2)) {
          if (value_is_indirect(val2)) {
            bank_diff = boolean_false;
          } else if (!sign) {
            bank_diff = boolean_false;
          } else {
            bank_diff = pic_are_variable_databits_different(sign, val2);
          }
        } 
        inc_op = (bank_diff) ? pic_opcode_incf : pic_opcode_incfsz;
        if (ii < value_sz_get(val2)) {
          pic_instr_append_f_d(pf, inc_op, val2, ii, pic_opdst_w);
        } else {
          pic_instr_append_f_d(pf, inc_op, sign, 0, pic_opdst_w);
        }
        if (bank_diff) {
          pic_instr_append_daop(pf, val1);
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
        }
      }
      if (ii < value_sz_get(val1)) {
        pic_indirect_bump(pf, val1, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_f_d(pf, pop_action_f, val1, ii, op_dst);
      } else {
        assert(sign);
        pic_instr_append_f_d(pf, pop_action_f, sign, 0, op_dst);
      }
      if (pic_opdst_w == op_dst) {
        pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
        pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
      }
    }
    if (sign) {
      pic_var_temp_release(pf, sign);
    }
  }
}

/* 
 * return the shift value for a multiply or divide,
 * or 0 if the value is not const or not shiftable
 * (assumes shift of 0 is optimized out earlier)
 */
static value_t pic_shift_get(pfile_t *pf, value_t val)
{
  value_t shift_val;

  shift_val = VALUE_NONE;
  if (value_is_const(val)) {
    variable_const_t n;

    n = value_const_get(val);
    if (!(n & (n - 1))) {
      /* make this a shift */
      variable_const_t shift;

      for (shift = -1; n; shift++, n >>= 1)
        ;
      shift_val = pfile_constant_get(pf, shift, 
          pic_result_vdef_get(val, VALUE_NONE, VALUE_NONE));
    }
  }
  return shift_val;
}

static void pic_op_mul(pfile_t *pf, operator_t op, value_t dst, value_t val1, 
    value_t val2)
{
  /* make sure the constant is on the right */
  value_t shift_val;

  if (value_is_const(val1)) {
    value_t tmp;

    tmp = val1;
    val1 = val2;
    val2 = tmp;
  }

  shift_val = pic_shift_get(pf, val2);
  if (shift_val) {
    pic_op(pf, operator_shift_left, dst, val1, shift_val);
    value_release(shift_val);
  } else {
    pic_var_mul_t mvars;
    variable_sz_t sz;

    sz = pic_result_sz_get(val1, val2, dst);

    pic_var_mul_get(pf, sz, &mvars);
    if ((value_is_same(val1, pic_last_values[PIC_LAST_VALUE_MULTIPLIER])
        && value_is_same(val2, pic_last_values[PIC_LAST_VALUE_MULTIPLICAND]))
        || (value_is_same(val1, pic_last_values[PIC_LAST_VALUE_MULTIPLICAND])
          && value_is_same(val2, pic_last_values[PIC_LAST_VALUE_MULTIPLIER]))
        ) {
      /* we needn't do anything here */
    } else {
      label_t fn;

      pic_op(pf, operator_assign, mvars.multiplier,   val1, VALUE_NONE);
      pic_op(pf, operator_assign, mvars.multiplicand, val2, VALUE_NONE);
      fn = pic_label_find(pf, PIC_LABEL_MULTIPLY, boolean_true);
      pic_instr_append_n(pf, pic_opcode_call, fn);
      label_release(fn);
      pic_last_values_reset();
      pic_last_value_set(PIC_LAST_VALUE_MULTIPLIER,   val1);
      pic_last_value_set(PIC_LAST_VALUE_MULTIPLICAND, val2);
    }
    pic_op(pf, operator_assign, dst, mvars.mresult, VALUE_NONE);
    pic_var_mul_release(pf, &mvars);
  }
}

static void pic_op_div_mod(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2)
{
  value_t shift_val;

  shift_val = pic_shift_get(pf, val2);
  if (shift_val) {
    if (operator_div == op) {
      /* right shift */
      pic_op(pf, operator_shift_right, dst, val1, shift_val);
    } else if (operator_mod == op) {
      /* binary and */
      value_t tmp;

      tmp = pfile_constant_get(pf, value_const_get(val2) - 1, 
          pic_result_vdef_get(val2, VALUE_NONE, VALUE_NONE));
      pic_op(pf, operator_andb, dst, val1, tmp);
      value_release(tmp);
    }
    value_release(shift_val);
  } else {
    variable_sz_t sz;
    pic_var_div_t dvars;

    sz = pic_result_sz_get(val1, val2, VALUE_NONE);
    pic_var_div_get(pf, sz, &dvars);
    if (value_is_same(pic_last_values[PIC_LAST_VALUE_DIVIDEND], val1)
        && value_is_same(pic_last_values[PIC_LAST_VALUE_DIVISOR], val2)) {
    } else {
      label_t fn;

      pic_op(pf, operator_assign, dvars.divisor,  val2, VALUE_NONE);
      pic_op(pf, operator_assign, dvars.dividend, val1, VALUE_NONE);
      fn = pic_label_find(pf,
          (VARIABLE_DEF_FLAG_SIGNED == pic_result_flag_get(val1, val2))
          ? PIC_LABEL_SDIVIDE
          : PIC_LABEL_DIVIDE,
          boolean_true);
      pic_instr_append_n(pf, pic_opcode_call, fn);
      label_release(fn);
      pic_last_values_reset();
      pic_last_value_set(PIC_LAST_VALUE_DIVIDEND, val1);
      pic_last_value_set(PIC_LAST_VALUE_DIVISOR, val2);
    }
    pic_op(pf, operator_assign, dst,
        (operator_div == op) ? dvars.quotient : dvars.remainder,
        VALUE_NONE);
    pic_var_div_release(pf, &dvars);
  }
}

/*
 *   signed:
 *      w    = val2_msb
 *      w   ^= 0x80
 *      temp = w
 *      w    = val1_msb
 *      w   ^= 0x80
 *      w    = temp - w
 *      goto both
 *   unsigned:
 *      w    = val1_msb
 *      w    = val2_msb - w
 *   both:
 *      btfsc STATUS, Z -- not equal, done!
 *      goto done
 *      w    = val1_lsb
 *      w    = val2_lsb - w
 *   done:
 *      Z    -- val1 == val2  (operator_eq, operator_ne)
 *      C,Z  -- val1 <= val2  (operator_le)
 *      C,NZ -- val1 < val2   (operator_lt)
 */
typedef enum {
  PIC_OP_RELATIONAL_TO_W_MOV,
  PIC_OP_RELATIONAL_TO_W_SUB,
  PIC_OP_RELATIONAL_TO_W_XOR
} pic_op_relational_to_w_t;

static void pic_op_relational_val_to_w(pfile_t *pf,
    pic_op_relational_to_w_t type, variable_sz_t ofs,
    value_t val, value_t val_sign, variable_const_t n)
{
  pic_opcode_t op_f;
  pic_opcode_t op_w;

  op_f = pic_opcode_movf;
  op_w = pic_opcode_movlw;
  switch (type) {
    case PIC_OP_RELATIONAL_TO_W_MOV:
      break;
    case PIC_OP_RELATIONAL_TO_W_SUB:
      op_f = pic_opcode_subwf;
      op_w = pic_opcode_sublw;
      break;
    case PIC_OP_RELATIONAL_TO_W_XOR:
      op_f = pic_opcode_xorwf;
      op_w = pic_opcode_xorlw;
      break;
  }
  if (ofs < value_sz_get(val)) {
    pic_instr_append_f_d(pf, op_f, val, ofs, pic_opdst_w);
  } else if (val_sign) {
    pic_instr_append_f_d(pf, op_f, val_sign, 0, pic_opdst_w);
  } else {
    pic_instr_append_w_kn(pf, op_w, n);
  }
}

operator_t pic_op_relational2(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2)
{
  size_t sz;

  sz = pic_result_sz_get(val1, val2, VALUE_NONE);
  if (value_is_const(val2)) {
    /* swap val2 & val1 */
    value_t tmp;

    tmp = val1;
    val1 = val2;
    val2 = tmp;
    /* invert the relation */
    if (operator_gt == op) {
      op = operator_lt;
    } else if (operator_ge == op) {
      op = operator_le;
    } else if (operator_le == op) {
      op = operator_ge;
    } else if (operator_lt == op) {
      op = operator_gt;
    }
  }

  if ((value_is_const(val1) && (sz > 1) && pic_result_is_signed(val1, val2))
      || value_is_lookup(val1)
      || value_is_bit(val1)
      || (value_is_indirect(val1) && value_is_indirect(val2))) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val1, val2, VALUE_NONE);
    op = pic_op_relational2(pf, op, dst, tmp, val2);
    pic_var_temp_release(pf, tmp);
  } else if ((value_is_const(val2) && (sz > 1))
      || value_is_lookup(val2)
      || value_is_bit(val2)) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val2, val1, VALUE_NONE);
    op = pic_op_relational2(pf, op, dst, val1, tmp);
    pic_var_temp_release(pf, tmp);
  } else {
    value_t          sign;
    boolean_t        is_first;
    variable_const_t n;

    if (value_is_const(val1)) {
      n = value_const_get(val1);
    }
    sign = VALUE_NONE;
    if (value_is_indirect(val1)) {
      pic_indirect_setup(pf, val1, value_sz_get(val1) - 1);
    } else if (value_is_indirect(val2)) {
      pic_indirect_setup(pf, val2, value_sz_get(val2) - 1);
    }
    if (value_sz_get(val1) < value_sz_get(val2)) {
      if (value_is_signed(val1)) {
        sign = pic_sign_get(pf, val1, value_sz_get(val1) - 1);
      }
    } else if (value_sz_get(val2) < value_sz_get(val1)) {
      sign = pic_sign_get(pf, val2, value_sz_get(val2) - 1);
    }
    is_first = boolean_true;
    if (pic_result_is_signed(val1, val2)) {
      value_t tmp;
      /*      w    = val2_msb
       *      w   ^= 0x80
       *      temp = w
       *      w    = val1_msb
       *      w   ^= 0x80
       *      w    = temp - w
       */      
      tmp = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
      if (value_is_const(val2)) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw,
            (value_const_get(val2) >> (8 * (sz - 1))) ^ 0x80);
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0x80);
        pic_op_relational_val_to_w(pf, PIC_OP_RELATIONAL_TO_W_XOR,
            sz - 1, val2, sign, 0); /* 0x00 ^ 0x80 */
      }
      pic_instr_append_f(pf, pic_opcode_movwf, tmp, 0);
      if (value_is_const(val1)) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw,
            (value_const_get(val1) >> (8 * (sz - 1))) ^ 0x80);
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0x80);
        pic_op_relational_val_to_w(pf, PIC_OP_RELATIONAL_TO_W_XOR,
            sz - 1, val1, sign, 0); /* 0x00 ^ 0x80 */
      }
      pic_instr_append_f_d(pf, pic_opcode_subwf, tmp, 0, pic_opdst_w);
      pic_var_temp_release(pf, tmp);
      sz--;
      is_first = boolean_false;
    } 
    if (sz) {
      label_t done;

      done = pfile_label_alloc(pf, 0);

      while (sz--) {
        if (!is_first) {
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_n(pf, pic_opcode_goto, done);
        }
        is_first = boolean_false;
        if (value_is_const(val1)) {
          pic_instr_append_w_kn(pf, pic_opcode_movlw, (n >> 8 * sz) & 0xff);
        } else {
          pic_op_relational_val_to_w(pf, PIC_OP_RELATIONAL_TO_W_MOV,
              sz, val1, sign, 0x00);
        }
        pic_op_relational_val_to_w(pf, PIC_OP_RELATIONAL_TO_W_SUB, 
            sz, val2, sign, 0x00);
      }
      pic_instr_append_label(pf, done);
      label_release(done);
    }
    if (sign) {
      pic_var_temp_release(pf, sign);
    }
    /* _c & !_z : val1 < val2  */
    /* _c & _z  : val1 <= val2 */
    /* !_c & _z : val1 >= val2 */
    /* !_c & !_z: val1 >  val2 */
    if (dst) {
      boolean_t single_bit;
      label_t   skip;

      if (value_is_indirect(dst)) {
        pic_indirect_setup(pf, dst, 0);
      }
      if (!value_is_indirect(dst)
        && !value_is_volatile(dst)
        && value_is_single_bit(dst)) {
        single_bit = boolean_true;
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
      } else {
        single_bit = boolean_false;
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
      }
      skip = ((operator_lt == op) || (operator_gt == op))
        ? pfile_label_alloc(pf, 0)
        : LABEL_NONE;
      switch (op) {
        case operator_lt:
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
          pic_instr_append_n(pf, pic_opcode_goto, skip);
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_c");
          break;
        case operator_le:
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_c");
          break;
        case operator_ge:
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_c");
          break;
        case operator_gt:
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
          pic_instr_append_n(pf, pic_opcode_goto, skip);
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_c");
          break;
        case operator_eq:
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          break;
        case operator_ne:
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
          break;
        default:
          break;
      }
      if (single_bit) {
        pic_instr_append_f(pf,pic_opcode_bsf, dst, 0);
        if (skip) {
          pic_instr_append_label(pf, skip);
        }
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
        if (skip) {
          pic_instr_append_label(pf, skip);
        }
        pic_set_to_w(pf, dst);
      }
      label_release(skip);
    }
  }
  return op;
}

static void pic_op_relational(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2)
{
  pic_op_relational2(pf, op, dst, val1, val2);
}

static void pic_op_equality(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2)
{
  assert(!(value_is_const(val1) && value_is_const(val2)));
  assert(!value_is_same(val1, val2));

  if ((value_is_indirect(val1) && value_is_indirect(val2))
      || value_is_bit(val1)
      || value_is_lookup(val1)) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val1, VALUE_NONE, VALUE_NONE);
    pic_op(pf, op, dst, tmp, val2);
    pic_var_temp_release(pf, tmp);
  } else if (value_is_single_bit(val1) 
      && value_is_const(val2)
      && ((0 == value_const_get(val2))
          || (1 == value_const_get(val2)))) {
    pic_opcode_t pop;

    pop = (operator_eq == op) ? pic_opcode_btfsc : pic_opcode_btfss;
    pic_instr_append(pf, pic_opcode_clrw);
    pic_instr_append_f(pf, pop, val1, 0);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
    pic_set_to_w(pf, dst);
  } else if ((!value_is_const(val2) && value_is_bit(val2))
      || value_is_lookup(val2)) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val2, VALUE_NONE, VALUE_NONE);
    pic_op(pf, op, dst, val1, tmp);
    pic_var_temp_release(pf, tmp);
  } else {
    value_t          tmp;
    variable_sz_t    sz_val1;
    variable_sz_t    sz_val2;
    variable_sz_t    sz;
    variable_sz_t    ii;
    variable_const_t n;
    value_t          sign;

    sz_val1 = value_byte_sz_get(val1);
    sz_val2 = value_byte_sz_get(val2);
    pic_indirect_setup3(pf, val1, val2, VALUE_NONE);
    sz = (sz_val1 > sz_val2) ? sz_val1 : sz_val2;
    if ((sz > 1) || value_is_indirect(dst)) {
      tmp = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
    } else {
      tmp = VALUE_NONE;
    }
    n = (value_is_const(val2)) 
      ? value_const_get(val2)
      : 0;
    sign = VALUE_NONE;
    for (ii = 0; ii < sz; ii++) {
      pic_indirect_bump3(pf, val1, val2, VALUE_NONE, ii, 
          PIC_INDIRECT_CHANGE_INCR);
      if ((ii == sz_val1) && value_is_signed(val1)) {
        assert(!sign);
        sign = pic_sign_get(pf, val1, ii - 1);
      } else if (ii == sz_val2) {
        if (value_is_signed(val2)) {
          assert(!sign);
          sign = pic_sign_get(pf, val2, ii - 1);
        } else {
          n = 0;
        }
      }
      if (ii >= sz_val1) {
        if (sign) {
          pic_instr_append_f_d(pf, pic_opcode_movf,  sign, 0, 
              pic_opdst_w);
        } else {
          pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
        }
      } else {
        pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii, pic_opdst_w);
      }
      if (value_is_const(val2) || ((ii >= value_sz_get(val2)) && !sign)) {
        /* when comparing one byte to 0 we needn't store anything! */
        unsigned n_lsb;

        n_lsb = n & 0xff;
        n >>= 8;
        if (n_lsb) {
          pic_instr_append_w_kn(pf, pic_opcode_sublw, n_lsb);
        }
      } else if (ii >= sz_val2) {
        pic_instr_append_f_d(pf, pic_opcode_subwf, sign, 0, pic_opdst_w);
      } else {
        pic_instr_append_f_d(pf, pic_opcode_subwf, val2, ii, pic_opdst_w);
      }
      if (tmp) {
        if (ii) {
          pic_instr_append_f_d(pf, pic_opcode_iorwf, tmp, 0, pic_opdst_f);
        } else {
          pic_instr_append_f(pf, pic_opcode_movwf, tmp, 0);
        }
      }
    }
    /* z is set if the values are equal, clear if not */
    if (dst) {
      boolean_t single_bit;

      if (value_is_single_bit(dst) 
          && !value_is_indirect(dst)
          && !value_is_volatile(dst)) {
        single_bit = boolean_true;
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
      } else {
        single_bit = boolean_false;
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
      }
      pic_instr_append_reg_flag(pf,
          (operator_eq == op) ? pic_opcode_btfsc : pic_opcode_btfss, 
          "_status", "_z");
      if (single_bit) {
        pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
        if (value_is_indirect(dst)) {
          pic_indirect_setup(pf, dst, 0);
        }
        pic_set_to_w(pf, dst);
      }
    }
    if (sign) {
      pic_var_temp_release(pf, sign);
    }
    if (tmp) {
      pic_var_temp_release(pf, tmp);
    }
  }
}

static void pic_op_andl_orl(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2)
{
  value_t      tmp;
  pic_opcode_t pop;

  assert(!value_is_const(val1));
  assert(!value_is_const(val2));
  if (operator_andl == op) {
    pop = pic_opcode_andwf;
  } else if (operator_orl == op) {
    pop = pic_opcode_iorwf;
  } else {
    assert(0);
  }
  tmp = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
  pic_op(pf, operator_logical, tmp, val2, VALUE_NONE);
  /* this will leave the result in w -- no reason to assign it anywhere */
  pic_op(pf, operator_logical, VALUE_NONE, val1, VALUE_NONE);
  pic_instr_append_f_d(pf, pop, tmp, 0, pic_opdst_w);
  if (value_is_single_bit(dst)) {
    pic_op_assign_single_bit_to_single_bit(pf, dst, VALUE_NONE, 
        PIC_ASSIGN_SINGLE_BIT_ACTION_NONE);
  } else if (value_is_bit(dst)) {
    pic_instr_append_f(pf, pic_opcode_movwf, tmp, 0);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
  } else {
    if (value_is_indirect(dst)) {
      pic_instr_append_f(pf, pic_opcode_movwf, tmp, 0);
      pic_indirect_setup(pf, dst, 0);
      pic_instr_append_f_d(pf, pic_opcode_movf, tmp, 0, pic_opdst_w);
    }
    pic_set_to_w(pf, dst);
  }
  pic_var_temp_release(pf, tmp);
}

static void pic_op_andb_orb_xorb_single_bit(pfile_t *pf, operator_t op, 
    value_t dst, value_t val1, value_t val2)
{
  if (value_is_volatile(val2)) {
    /* swap val2 & val1 */
    value_t tmp;

    tmp  = val1;
    val1 = val2;
    val2 = tmp;
  }

  switch (op) {
    case operator_andb:
      if (!value_is_single_bit(dst)) {
        /* simple case 1 : set w to 0 or 1, then assign to dst */
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
        pic_instr_append_f(pf, pic_opcode_btfsc, val1, 0);
        pic_instr_append_f(pf, pic_opcode_btfss, val2, 0);
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
        pic_set_to_w(pf, dst);
      } else if (!value_is_volatile(val2)
        && (value_is_same(dst, val1)
          || value_is_same(dst, val2))) {
        /* this cannot be used when val2 is volatile because it requires 
           accessing val2 twice; both val1 and dst can be volatile though */
        label_t clr;

        clr = pfile_label_alloc(pf, 0);
        pic_instr_append_f(pf, pic_opcode_btfss, val1, 0);
        pic_instr_append_n(pf, pic_opcode_goto, clr);
        pic_instr_append_f(pf, pic_opcode_btfsc, val2, 0);
        pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
        pic_instr_append_f(pf, pic_opcode_btfss, val2, 0);
        pic_instr_append_label(pf, clr);
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
        label_release(clr);
      } else if (value_is_volatile(dst)) {
        /* this is probably cheezy, but I'm going to use _status:_z as a
           temporary */
        pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_z");
        pic_instr_append_f(pf, pic_opcode_btfsc, val1, 0);
        pic_instr_append_f(pf, pic_opcode_btfss, val2, 0);
        pic_instr_append_reg_flag(pf, pic_opcode_bsf, "_status", "_z");
        pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
        pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
        pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
      } else {
        /* simple case 2 : dst cannot be volatile as it's set twice */
        pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
        pic_instr_append_f(pf, pic_opcode_btfsc, val1, 0);
        pic_instr_append_f(pf, pic_opcode_btfss, val2, 0);
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
      }
      break;
    case operator_orb:
      if (!value_is_single_bit(dst)) {
        /* simple case 1 : set w to 0 or 1, then assign to dst */
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
        pic_instr_append_f(pf, pic_opcode_btfss, val1, 0);
        pic_instr_append_f(pf, pic_opcode_btfsc, val2, 0);
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
        pic_set_to_w(pf, dst);
      } else if (!value_is_volatile(val2)
        && (value_is_same(dst, val1)
          || value_is_same(dst, val2))) {
        /* this cannot be used when val2 is volatile because it requires 
           accessing val2 twice; both val1 and dst can be volatile though */
        label_t set;

        set = pfile_label_alloc(pf, 0);
        pic_instr_append_f(pf, pic_opcode_btfsc, val1, 0);
        pic_instr_append_n(pf, pic_opcode_goto, set);
        pic_instr_append_f(pf, pic_opcode_btfss, val2, 0);
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
        pic_instr_append_f(pf, pic_opcode_btfsc, val2, 0);
        pic_instr_append_label(pf, set);
        pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
        label_release(set);
      } else if (value_is_volatile(dst)) {
        /* this is probably cheezy, but I'm going to use _status:_z as a
           temporary */
        pic_instr_append_reg_flag(pf, pic_opcode_bsf, "_status", "_z");
        pic_instr_append_f(pf, pic_opcode_btfss, val1, 0);
        pic_instr_append_f(pf, pic_opcode_btfsc, val2, 0);
        pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_z");
        pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
        pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
        pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
      } else {
        /* simple case 2 : dst cannot be volatile as it's set twice */
        pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
        pic_instr_append_f(pf, pic_opcode_btfss, val1, 0);
        pic_instr_append_f(pf, pic_opcode_btfsc, val2, 0);
        pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
      }
      break;
    case operator_xorb:
      {
        label_t set;
        label_t clr;
        label_t skip;
        label_t done;

        done = pfile_label_alloc(pf, 0);
        skip = pfile_label_alloc(pf, 0);
        clr  = pfile_label_alloc(pf, 0);

        if (!value_is_volatile(dst) 
            && value_is_single_bit(dst)
            && !value_is_same(dst, val1)
            && !value_is_same(dst, val2)) {
          /* we save *one* byte in this case, woo woo! */
          if (value_is_single_bit(dst)) {
            pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
          } else {
            pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
          }
          set = done;
        } else {
          set = pfile_label_alloc(pf, 0);
        }

        pic_instr_append_f(pf, pic_opcode_btfss, val1, 0);
        pic_instr_append_n(pf, pic_opcode_goto, skip);
        /* v1 = 1 */
        pic_instr_append_f(pf, pic_opcode_btfsc, val2, 0);
        pic_instr_append_n(pf, pic_opcode_goto, clr); /* v1=1, v2=1 */
        pic_instr_append_n(pf, pic_opcode_goto, set); /* v1=1, v2=0 */
        /* v1 = 0 */
        pic_instr_append_label(pf, skip);
        pic_instr_append_f(pf, pic_opcode_btfsc, val2, 0);
        pic_instr_append_n(pf, pic_opcode_goto, set); /* v1=0, v2=0 */
        pic_instr_append_label(pf, clr);
        if (value_is_single_bit(dst)) {
          pic_instr_append_f(pf, pic_opcode_bcf, dst, 0);
        } else {
          pic_instr_append(pf, pic_opcode_clrw);
        }
        if (set != done) {
          pic_instr_append_n(pf, pic_opcode_goto, done);
          pic_instr_append_label(pf, set);
          if (value_is_single_bit(dst)) {
            pic_instr_append_f(pf, pic_opcode_bsf, dst, 0);
          } else {
            pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
          }
        }
        pic_instr_append_label(pf, done);
        if (!value_is_single_bit(dst)) {
          pic_set_to_w(pf, dst);
        }

        if (set != done) {
          label_release(set);
        }
        label_release(clr);
        label_release(skip);
        label_release(done);
      }
      break;
    default:
      assert(0);
  }
}

static void pic_op_andb_orb_xorb(pfile_t *pf, operator_t op, value_t dst,
    value_t val1, value_t val2)
{
  /*
   * only one of (dst, val1, val2) can be indirect!
   * also, I've no intention of dealing with bit values here
   * -- make them something useful first
   */
  if (value_is_single_bit(val1) && value_is_single_bit(val2)) {
    pic_op_andb_orb_xorb_single_bit(pf, op, dst, val1, val2);
  } else if (value_is_bit(val2)
      || value_is_lookup(val2)
      || (value_is_indirect(val2) 
        && (value_is_indirect(dst) || value_is_indirect(val1)))) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val2, VALUE_NONE, dst);
    pic_op(pf, op, dst, val1, tmp);
    pic_var_temp_release(pf, tmp);
  } else if (value_is_bit(val1)
      || value_is_lookup(val1)
      || (value_is_indirect(dst) && value_is_indirect(val1))) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val1, VALUE_NONE, dst);
    pic_op(pf, op, dst, tmp, val2);
    pic_var_temp_release(pf, tmp);
  } else if (value_is_bit(dst)) {
    value_t tmp;

    tmp = pic_var_temp_get_def(pf, value_def_get(dst));
    pic_op(pf, op, tmp, val1, val2);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else {
    /* do the work! */
    pic_opcode_t     pop;
    variable_const_t n;
    variable_sz_t    dst_sz;
    variable_sz_t    val1_sz;
    variable_sz_t    val2_sz;
    variable_sz_t    sz;
    variable_sz_t    ii;
    value_t          sign;

    n = 0;
    if (value_is_same(dst, val1)
      || value_is_const(val2)) {
      /* 
       * if dst & val1 are the same, swap val1 with val2
       * the general bit is:
       *    movf val1, w
       *    pic_op val2,[w|f]
       *    [movwf dst]
       * so we save an instructions if val2 and dst are the same
       *
       * if val2 is const, move it to val1 to simplify things
       */
      value_t tmp;

      tmp = val1;
      val1 = val2;
      val2 = tmp;
    }
    
    if (value_is_const(val1)) {
      n = value_const_get(val1);
    }
    dst_sz  = value_sz_get(dst);
    val1_sz = value_sz_get(val1);
    val2_sz = value_sz_get(val2);
    sz = pic_result_sz_get(val1, val2, dst);
    /* 
     * once we run out of data for val1 or val2 special handling
     * kicks in
     */
    if ((sz > value_sz_get(val1)) && !value_is_signed(val1)) {
      sz = value_sz_get(val1);
    } else if ((sz > value_sz_get(val2)) && !value_is_signed(val2)) {
      sz = value_sz_get(val2);
    }
    pic_indirect_setup3(pf, val1, val2, dst);
    sign = VALUE_NONE;
    for (ii = 0; ii < sz; ii++) {
      boolean_t skip_dst;

      if (operator_andb == op) {
        pop = pic_opcode_andwf;
      } else if (operator_orb == op) {
        pop = pic_opcode_iorwf;
      } else if (operator_xorb == op) {
        pop = pic_opcode_xorwf;
      } else {
        assert(0);
      }

      if ((ii == val1_sz) 
        && value_is_signed(val1) 
        && !value_is_const(val1)) {
        sign = pic_sign_get(pf, val1, ii - 1);
      }
      if (ii == val2_sz) {
        sign = pic_sign_get(pf, val2, ii - 1);
      }
      pic_indirect_bump3(pf, val1, val2, dst, ii, PIC_INDIRECT_CHANGE_INCR);
      skip_dst = boolean_false;
      if (ii < val1_sz) {
        if (value_is_const(val1)) {
          unsigned char n_lsb;

          n_lsb = n & 0xff;
          n >>= 8;
          /* what we do here depends on n and op */
          if (0x00 == n_lsb) {
            skip_dst = boolean_true;
            if (operator_andb == op) {
              /* 0 & x = 0 */
              pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
            } else if ((operator_orb == op) 
              || (operator_xorb == op)) {
              /* 0 | x == 0 ^ x == x */
            } else {
              assert(0);
            }
          } else if (0xff == n_lsb) {
            if (operator_andb == op) {
              /* 0xff & x = x */
              /*skip_dst = boolean_true;*/
              pop = pic_opcode_movf;
            } else if (operator_orb == op) {
              /* 0xff | x = 0xff */
              pic_instr_append_w_kn(pf, pic_opcode_movlw, 0xff);
              pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
              skip_dst = boolean_true;
            } else if (operator_xorb == op) {
              /* 0xff ^ x = ~x */
              /* pic_instr_append_w_kn(pf, pic_opcode_movlw, n_lsb); */
              pop = pic_opcode_comf;
            }
          } else {
            pic_instr_append_w_kn(pf, pic_opcode_movlw, n_lsb);
          }
        } else {
          pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii, pic_opdst_w);
        }
      } else if (sign) {
        pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
      } else {
        assert(0);
      }
      if (!skip_dst) {
        pic_opdst_t op_dst;

        op_dst = (value_is_same(val2, dst))
          ? pic_opdst_f
          : pic_opdst_w;
        if (ii < val2_sz) {
          pic_instr_append_f_d(pf, pop, val2, ii, op_dst);
        } else if (sign) {
          pic_instr_append_f_d(pf, pop, sign, 0, op_dst);
        } else {
          assert(0);
        }
        if (pic_opdst_w == op_dst) {
          pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
        }
      }
    }
    if ((sz < val1_sz) && (sz < dst_sz)) {
      /* 
       * val1 is now all zeros.
       * if this is AND --> zero out dst
       * if this is OR or XOR 
       *   if val2 != dst
       *     copy val2 to dst
       *   if val2 == dst *and* dst is indirect *and* 
       *     the result is signed *and* sizeof(dst) > sz
       *     increment to the end of dst for sign extension
       */
       if (operator_andb == op) {
         /* zero out the remainder */
         while ((ii < val1_sz) && (ii < dst_sz)) {
           pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
           pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
           ii++;
         }
       } else if (!value_is_same(val2, dst)) {
         /* copy val2 to dst */
         while ((ii < val1_sz) && (ii < dst_sz)) {
           pic_indirect_bump3(pf, VALUE_NONE, val1, dst, ii,
             PIC_INDIRECT_CHANGE_INCR);
           pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii, pic_opdst_w);
           pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
           ii++;
         }
       } else if ((sz < dst_sz) && value_is_indirect(dst)) {
         /* we need sign extension on dst, so increment as necessary */
         if (sz + 1 == value_sz_get(dst)) {
           pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
         } else {
           pic_instr_append_w_kn(pf, pic_opcode_movlw, value_sz_get(dst) - sz);
           pic_instr_append_reg_d(pf, pic_opcode_addwf, "_fsr", pic_opdst_f);
         }
       }
    } else if ((sz < val2_sz) && (sz < dst_sz)) {
      /*
       * val2 is now all zeros
       * if this is AND --> zero out dst
       * if this is OR or XOR
       *   copy the remainder of val1 to dst
       */
      while ((ii < val2_sz) && (ii < dst_sz))  {
        if (operator_andb == op) {
          pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
          pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
        } else {
          pic_indirect_bump3(pf, val1, VALUE_NONE, dst, ii, 
            PIC_INDIRECT_CHANGE_INCR);
          if (ii < val1_sz) {
            pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii, pic_opdst_w);
            pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
          }
        }
        ii++;
      }
    }
    pic_sign_extend(pf, pic_result_is_signed(val1, val2), dst, ii);
  }
}

/* this does a single shift + assign */
static void pic_op_shift_assign(pfile_t *pf, operator_t op, value_t dst,
    value_t src)
{
  variable_sz_t sz;
  variable_sz_t dst_sz;
  variable_sz_t src_sz;
  pic_opdst_t   op_dst;

  assert(!(value_is_indirect(dst) && value_is_indirect(src)));
  assert(!value_is_bit(dst));
  assert(!value_is_bit(src));

  src_sz = value_sz_get(src);
  dst_sz = value_sz_get(dst);
  sz     = (src_sz < dst_sz) ? src_sz : dst_sz;
  pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_c");
  if (operator_shift_left == op) {
    variable_sz_t ii;

    if (value_is_indirect(src)) {
      pic_indirect_setup(pf, src, 0);
    } else if (value_is_indirect(dst)) {
      pic_indirect_setup(pf, dst, 0);
    }
    op_dst = (value_is_same(dst, src))
      ? pic_opdst_f
      : pic_opdst_w;

    for (ii = 0; ii < sz; ii++) {
      pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
      if (!value_is_same(dst, src)) {
        pic_indirect_bump(pf, src, ii, PIC_INDIRECT_CHANGE_INCR);
      }
      pic_instr_append_f_d(pf, pic_opcode_rlf, src, ii, op_dst);
      if (pic_opdst_w == op_dst) {
        pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
      }
    }
    if (sz < dst_sz) {
      pic_sign_extend(pf, value_is_signed(src), dst, ii);
    }
  } else if (operator_shift_right == op) {
    variable_sz_t ii;

    if (value_is_indirect(src)) {
      pic_indirect_setup(pf, src, sz - 1);
    } else if (value_is_indirect(dst)) {
      pic_indirect_setup(pf, dst, sz - 1);
    }
    op_dst = (value_is_same(dst, src))
      ? pic_opdst_f
      : pic_opdst_w;
    if (value_is_signed(src)) {
      /* signed shift == arithmetic */
      pic_instr_append_f_bn(pf, pic_opcode_btfsc, src, sz - 1, 7);
      pic_instr_append_reg_flag(pf, pic_opcode_bsf, "_status", "_c");
    }
    for (ii = sz; ii; ii--) {
      pic_indirect_bump(pf, dst, (ii != sz), PIC_INDIRECT_CHANGE_DECR);
      if (!value_is_same(dst, src)) {
        pic_indirect_bump(pf, src, (ii != sz), PIC_INDIRECT_CHANGE_DECR);
      }
      pic_instr_append_f_d(pf, pic_opcode_rrf, src, ii - 1, op_dst);
      if (pic_opdst_w == op_dst) {
        pic_instr_append_f(pf, pic_opcode_movwf, dst, ii - 1);
      }
    }
    if (sz < dst_sz) {
      if (value_is_indirect(dst)) {
        pic_indirect_setup(pf, dst, sz - 1);
      }
      pic_sign_extend(pf, value_is_signed(src), dst, ii);
    }
  } else {
    assert(0);
  }
}

static void pic_set_to_zero(pfile_t *pf, value_t dst)
{
  if (value_is_bit(dst)) {
    value_t n;

    n = pfile_constant_get(pf, 0, VARIABLE_DEF_NONE);
    pic_op(pf, operator_assign, dst, n, VALUE_NONE);
    value_release(n);
  } else {
    variable_sz_t ii;

    if (value_is_indirect(dst)) {
      pic_indirect_setup(pf, dst, 0);
    }
    for (ii = 0; ii < value_sz_get(dst); ii++) {
      if (ii && value_is_indirect(dst)) {
        pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
      }
      pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
    }
  }
}

/*
 * this can be optimized somewhat because val2 is const;
 * all we really do here is inline shift unless val2
 * is larger than some constant (tbd)
 *
 * at this point we know:
 *   dst is not volatile and not bit
 *   val1 is not bit, lookup, or const
 *   val1 & dst are not *both* indirect
 */
static void pic_op_shift_left_const(pfile_t *pf, value_t dst, value_t val1,
    value_t val2) 
{
  variable_sz_t    sz;
  variable_const_t shift;
  variable_const_t byte_shift; /* # of bytes to shift */
  variable_const_t bit_shift;  /* # of bits to shift  */

  shift = value_const_get(val2);
  byte_shift = shift / 8;
  bit_shift  = shift % 8;

  sz = pic_result_sz_get(val1, val2, dst);
  if (shift >= 8 * sz) {
    pic_set_to_zero(pf, dst);
  } else {
    assert(sz > byte_shift);
    if (value_is_indirect(dst) && value_is_same(dst, val1)) {
      /* there's no optimization that can be done here */
      bit_shift += byte_shift * 8;
      byte_shift = 0;
    } else if (byte_shift) {
      variable_sz_t ii;
      /*
       * this needs to be done from back to front in case dst & val1
       * are the same!
       */ 
      if (value_is_indirect(dst)) {
        pic_indirect_setup(pf, dst, sz - 1);
      } else if (value_is_indirect(val1)) {
        pic_indirect_setup(pf, val1, sz - 1 - byte_shift);
      }
      for (ii = sz; ii > byte_shift; ii--) {
        if (ii != sz) {
          pic_indirect_bump3(pf, dst, val1, VALUE_NONE, ii, 
              PIC_INDIRECT_CHANGE_DECR);
        }
        if (ii - 1 - byte_shift >= value_sz_get(val1)) {
          pic_instr_append_f(pf, pic_opcode_clrf, dst, ii - 1);
        } else {
          pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii - 1 - byte_shift,
              pic_opdst_w);
          pic_instr_append_f(pf, pic_opcode_movwf, dst, ii - 1);
        }
      }
      while (ii--) {
        pic_indirect_bump(pf, dst, 1, PIC_INDIRECT_CHANGE_DECR);
        pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
      }
      if (!bit_shift 
          && pic_result_is_signed(val1, val2)
          && value_is_indirect(dst)
          && (sz < value_sz_get(dst))) {
        /* position dst back to the end for sign extension */
        pic_indirect_setup(pf, dst, sz - 1);
      }
    }
    if (bit_shift) {
      if (!byte_shift && !value_is_same(dst, val1)) {
        pic_op_shift_assign(pf, operator_shift_left, dst, val1);
        bit_shift--;
      }
      if (bit_shift) {
        value_t       tmp;
        label_t       top;
        variable_sz_t ii;

        tmp = VALUE_NONE;
        top = LABEL_NONE;
        if (bit_shift > 1) {
          tmp = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
          pic_instr_append_w_kn(pf, pic_opcode_movlw, bit_shift);
          pic_instr_append_f(pf, pic_opcode_movwf, tmp, 0);
          top = pfile_label_alloc(pf, 0);
          pic_instr_append_label(pf, top);
        }
        if (value_is_indirect(dst)) {
          pic_indirect_setup(pf, dst, 0);
        }
        pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_c");
        for (ii = 0; ii < sz; ii++) {
          pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
          pic_instr_append_f_d(pf, pic_opcode_rlf, dst, ii, pic_opdst_f);
        }
        if (tmp) {
          pic_instr_append_f_d(pf, pic_opcode_decfsz, tmp, 0, pic_opdst_f);
          pic_instr_append_n(pf, pic_opcode_goto, top);
          pic_var_temp_release(pf, tmp);
          label_release(top);
        }
      }
    }
    pic_sign_extend(pf, pic_result_is_signed(val1, val2), dst, sz);
  }
}

/*
 * this can be optimized somewhat because val2 is const;
 * all we really do here is inline shift unless val2
 * is larger than some constant (tbd)
 *
 * at this point we know:
 *   dst is not volatile and not bit
 *   val1 is not bit, lookup, or const
 *   val1 & dst are not *both* indirect
 */
static void pic_op_shift_right_const(pfile_t *pf, value_t dst, value_t val1, 
    value_t val2)
{
  size_t           sz;
  variable_const_t shift;
  variable_const_t byte_shift; /* # of bytes to shift */
  variable_const_t bit_shift;  /* # of bits to shift  */

  shift = value_const_get(val2);
  byte_shift = shift / 8;
  bit_shift  = shift % 8;

  if (shift >= 8 * value_sz_get(val1)) {
    pic_set_to_zero(pf, dst);
  } else if (value_sz_get(dst) < value_sz_get(val1)) {
    /* do the work in a temporary */
    value_t tmp;

    tmp = pic_var_temp_get(pf, 
        (value_is_signed(val1)) 
        ? VARIABLE_DEF_FLAG_SIGNED : VARIABLE_DEF_FLAG_NONE,
        value_sz_get(val1));
    pic_op_shift_right_const(pf, tmp, val1, val2);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else {
    sz = pic_result_sz_get(val1, val2, dst);
    assert(sz > byte_shift);
    if (value_is_indirect(dst) && value_is_same(dst, val1)) {
      bit_shift += byte_shift * 8;
      byte_shift = 0;
    } else if (byte_shift) {
      variable_sz_t ii;
      value_t       sign;

      if (value_is_indirect(dst)) {
        pic_indirect_setup(pf, dst, 0);
      } else if (value_is_indirect(val2)) {
        pic_indirect_setup(pf, val1, byte_shift);
      }
      sign = (value_is_signed(val1))
        ? pic_sign_get(pf, val1, value_sz_get(val1) - 1)
        : VALUE_NONE;

      for (ii = 0; ii < byte_shift; ii++) {
        pic_indirect_bump3(pf, dst, val1, VALUE_NONE, ii,
            PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii + byte_shift,
            pic_opdst_w);
        pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
      }
      if (sign) {
        pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
      }
      while (ii < sz) {
        pic_indirect_bump(pf, dst, ii, PIC_INDIRECT_CHANGE_INCR);
        if (sign) {
          pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
        } else {
          pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
        }
        ii++;
      }
      if (sign) {
        pic_var_temp_release(pf, sign);
      }
    }
    if (bit_shift) {
      if (!byte_shift && !value_is_same(dst, val1)) {
        pic_op_shift_assign(pf, operator_shift_right, dst, val1);
        bit_shift--;
      }
      if (bit_shift) {
        value_t       tmp;
        label_t       top;
        variable_sz_t ii;

        tmp = VALUE_NONE;
        top = LABEL_NONE;
        if (bit_shift > 1) {
          tmp = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
          pic_instr_append_w_kn(pf, pic_opcode_movlw, bit_shift);
          pic_instr_append_f(pf, pic_opcode_movwf, tmp, 0);
          top = pfile_label_alloc(pf, 0);
          pic_instr_append_label(pf, top);
        }
        if (value_is_indirect(dst)) {
          pic_indirect_setup(pf, dst, sz - 1);
        }
        pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_c");
        if (value_is_signed(dst)) {
          pic_instr_append_f_bn(pf, pic_opcode_btfsc, dst, sz - 1, 7);
          pic_instr_append_reg_flag(pf, pic_opcode_bsf, "_status", "_c");
        }
        for (ii = sz; ii; ii--) {
          if (ii != sz) {
            pic_indirect_bump(pf, dst, ii - 1, PIC_INDIRECT_CHANGE_DECR);
          }
          pic_instr_append_f_d(pf, pic_opcode_rrf, dst, ii - 1, pic_opdst_f);
        }
        if (tmp) {
          pic_instr_append_f_d(pf, pic_opcode_decfsz, tmp, 0, pic_opdst_f);
          pic_instr_append_n(pf, pic_opcode_goto, top);
          pic_var_temp_release(pf, tmp);
          label_release(top);
        }
      }
    }
    /*
     * there's no way the sign bit can be 1 after a right shift, so
     * just pretend the result was unsigned
     */
    pic_sign_extend(pf, boolean_false, dst, sz);
  }
}

static void pic_op_shift_const(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2)
{
  variable_const_t n;

  n = value_const_get(val2) & (8UL * value_sz_get(dst) - 1);
  if (!n) {
    pic_set_to_zero(pf, dst);
  } else {
    if (operator_shift_left == op) {
      pic_op_shift_left_const(pf, dst, val1, val2);
    } else {
      pic_op_shift_right_const(pf, dst, val1, val2);
    }
  }
}

/* possibilities:
 *   val2 == const --> either inline or loop, depends on val2
 *   dst  == val1; dst != volatile --> do it
 *   dst  != val1 
 *     dst volatile --> shift val1 into tmp; work on tmp; assign tmp to dst
 *     dst not volatile --> shift val1 into dst; work on dst
 */
static void pic_op_shift(pfile_t *pf, operator_t op, value_t dst, value_t val1, 
value_t val2)
{
  if (value_is_bit(val1) 
      || value_is_lookup(val1) 
      || value_is_const(val1)
      || (!value_is_same(dst, val1)
        && value_is_indirect(dst)
        && value_is_indirect(val1))) {
    value_t tmp;

    tmp = pic_var_temp_get_assign(pf, val1, VALUE_NONE, dst);
    pic_op(pf, op, dst, tmp, val2);
    pic_var_temp_release(pf, tmp);
  } else if (value_is_bit(dst) || value_is_volatile(dst)) {
    value_t tmp;

    tmp = pic_var_temp_get_def(pf, value_def_get(val1));
    pic_op(pf, op, tmp, val1, val2);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else if (value_is_const(val2)) {
    pic_op_shift_const(pf, op, dst, val1, val2);
  } else {
    variable_sz_t sz;
    variable_sz_t val1_sz;
    variable_sz_t dst_sz;
    label_t       loop_top;
    label_t       loop_decr;
    label_t       loop_done;
    value_t       loop_var;
    value_t       work_var;
    value_t       cn;
    boolean_t     def_changed;

    loop_top  = pfile_label_alloc(pf, 0);
    loop_decr = pfile_label_alloc(pf, 0);
    loop_done = pfile_label_alloc(pf, 0);

    loop_var  = pic_var_temp_get(pf, VARIABLE_DEF_FLAG_NONE, 1);
    cn        = pfile_constant_get(pf, value_sz_get(val1) * 8 - 1, 
                  VARIABLE_DEF_NONE);
    val1_sz = value_sz_get(val1);
    dst_sz  = value_sz_get(dst);

    pic_op(pf, operator_andb, loop_var, val2, cn);
    pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
    pic_instr_append_n(pf, pic_opcode_goto, loop_done);
    /* do we need to work in a temporary? */
    def_changed = boolean_false;
    if (value_is_bit(dst)
      || (value_is_volatile(dst)
      || (!value_is_same(dst, val1)
        && value_is_indirect(dst) 
        && value_is_indirect(val1)))) {
      /* tmp = val1 */
      work_var = pic_var_temp_get_assign(pf, val1, val2, dst);
      pic_instr_append_n(pf, pic_opcode_goto, loop_decr);
    } else {
      if (value_sz_get(val1) < value_sz_get(dst)) {
        work_var = value_clone(dst);
        value_def_set(work_var, pic_result_vdef_get(val1, val2, VALUE_NONE));
        def_changed = boolean_true;
      } else {
        work_var = dst;
      }
      if (!value_is_same(work_var, val1)) {
        pic_op_shift_assign(pf, op, work_var, val1);
        pic_instr_append_f_d(pf, pic_opcode_decf, loop_var, 0, pic_opdst_f);
        pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
        pic_instr_append_n(pf, pic_opcode_goto, loop_done);
      }
    }
    pic_instr_append_label(pf, loop_top);
    pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_c");
    sz = value_sz_get(work_var);
    if (operator_shift_left == op) {
      variable_sz_t ii;

      if (value_is_indirect(work_var)) {
        pic_indirect_setup(pf, work_var, 0);
      } else if (value_is_indirect(val1)) {
        pic_indirect_setup(pf, val1, 0);
      }
      for (ii = 0; ii < sz; ii++) {
        pic_indirect_bump(pf, work_var, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_indirect_bump(pf, val1, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_f_d(pf, pic_opcode_rlf, work_var, ii,
            pic_opdst_f);
      }
    } else if (operator_shift_right == op) {
      variable_sz_t ii;

      if (value_is_indirect(work_var)) {
        pic_indirect_setup(pf, work_var, sz - 1);
      } else if (value_is_indirect(val1)) {
        pic_indirect_setup(pf, val1, sz - 1);
      }
      if (value_is_signed(val1)) {
        pic_instr_append_f_bn(pf, pic_opcode_btfsc, val1, sz - 1, 7);
        pic_instr_append_reg_flag(pf, pic_opcode_bsf, "_status", "_c");
      }
      for (ii = sz; ii; ii--) {
        pic_indirect_bump(pf, work_var, ii, PIC_INDIRECT_CHANGE_DECR);
        pic_indirect_bump(pf, val1, ii, PIC_INDIRECT_CHANGE_DECR);
        if (value_is_same(work_var, val1)) {
          pic_instr_append_f_d(pf, pic_opcode_rrf, work_var, ii - 1, 
              pic_opdst_f);
        } else {
          pic_instr_append_f_d(pf, pic_opcode_rrf, val1, ii - 1,
              pic_opdst_w);
          pic_instr_append_f(pf, pic_opcode_movwf, work_var, ii - 1);
        }
      }
    } else {
      assert(0);
    }
    pic_instr_append_label(pf, loop_decr);
    pic_instr_append_f_d(pf, pic_opcode_decfsz, loop_var, 0, pic_opdst_f);
    pic_instr_append_n(pf, pic_opcode_goto, loop_top);
    if (!def_changed && (work_var != dst)) {
      pic_op_shift_assign(pf, op, dst, work_var);
    }
    pic_instr_append_label(pf, loop_done);
    label_release(loop_done);
    label_release(loop_decr);
    label_release(loop_top);
    if (value_is_same(work_var, dst) || def_changed) {
      if (sz < value_sz_get(dst)) {
        if (value_is_indirect(dst) && (operator_shift_right == op)) {
          pic_indirect_setup(pf, dst, sz - 1);
        }
        pic_sign_extend(pf, pic_result_is_signed(val1, val2), dst, sz);
      }
      work_var = 0;
    }
    if (work_var) {
      pic_var_temp_release(pf, work_var);
    }
    pic_var_temp_release(pf, loop_var);
  }
}

/* 
 * a trivial increment/decrement (one byte src --> dst,
 * this is optimized somewhat, otherwise it's
 * dst = src; dst++
 */
static void pic_op_incr_decr(pfile_t *pf, operator_t op, value_t dst, 
    value_t val, value_t dummy)
{
  variable_sz_t sz;

  assert(!value_is_const(val));
  sz = pic_result_sz_get(val, VALUE_NONE, dst);
  if ((sz == 1) 
      && !value_is_bit(dst)
      && !value_is_bit(val) 
      && !value_is_lookup(val)
      && !(value_is_indirect(val) && value_is_indirect(dst))
      && !value_is_same(dst, val)) {
    /*
     * this is the trivial case
     *   x +/- 1 --> y
     */
    pic_indirect_setup3(pf, dst, val, VALUE_NONE);
    pic_instr_append_f_d(pf, 
        (operator_incr == op) ? pic_opcode_incf : pic_opcode_decf,
        val, 0, pic_opdst_w);
    if (value_variable_get(dst)) {
      pic_instr_append_f(pf, pic_opcode_movwf, dst, 0);
      pic_sign_extend(pf, value_is_signed(val), dst, sz);
    }
  } else if (value_is_bit(dst)
      || (value_is_volatile(dst) && !value_is_same(dst, val))) {
    /* val --> tmp; tmp++; tmp --> dst */
    value_t tmp;

    tmp = pic_var_temp_get(pf, pic_result_flag_get(val, VALUE_NONE), sz);
    pic_op(pf, operator_assign, tmp, val, VALUE_NONE);
    pic_op(pf, op, tmp, tmp, VALUE_NONE);
    pic_op(pf, operator_assign, dst, tmp, VALUE_NONE);
    pic_var_temp_release(pf, tmp);
  } else {
    variable_sz_t  ii;
    const char    *flag;
    pic_opcode_t   pop;
    pic_opcode_t   pop2;
    value_t        work_val;
    variable_def_t work_def;

    work_val = dst;
    work_def = VARIABLE_DEF_NONE;
    if (!value_is_same(dst, val)) {
      if (value_sz_get(dst) > sz) {
        work_def = variable_def_alloc(0, variable_def_type_integer, 
            VARIABLE_DEF_FLAG_NONE, sz);
        work_val = value_clone(work_val);
        value_def_set(work_val, work_def);
      }
      pic_op(pf, operator_assign, work_val, val, VALUE_NONE);
    }

    if ((sz > 1) && 
        ((operator_decr == op)
        || value_is_indirect(work_val))) {
      if (value_is_indirect(work_val)) {
        pic_indirect_setup(pf, work_val, 0);
      }
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      flag = "_c";
      if (operator_incr == op) {
        pop = pic_opcode_addwf;
        pop2 = pic_opcode_btfsc;
      } else {
        pop = pic_opcode_subwf;
        pop2 = pic_opcode_btfss;
      }
    } else {
      flag = "_z";
      pop  = (operator_incr == op) ? pic_opcode_incf : pic_opcode_decf;
      pop2 = pic_opcode_btfsc;
    }
    for (ii = 0; ii < sz; ii++) {
      if (ii) {
        pic_indirect_bump(pf, work_val, ii, PIC_INDIRECT_CHANGE_INCR);
        pic_instr_append_reg_flag(pf, pop2, "_status", flag);
      }
      pic_instr_append_f_d(pf, pop, work_val, ii, pic_opdst_f);
    }
    if (work_def) {
      pic_sign_extend(pf, value_is_signed(val), dst, sz);
    }
    if (work_val != dst) {
      value_release(work_val);
    }
  }
}

/*
 * NAME
 *   pic_op
 *
 * DESCRIPTION
 *   create code for an operator
 *
 * PARAMETERS
 *   pf  : file
 *   op  :
 *   dst :
 *   val1:
 *   val2:
 *
 * RETURN
 *   none
 *
 * NOTES
 */
extern variable_sz_t pic_temp_sz_in_use;

typedef void (*operator_fn_t)(pfile_t *pf, operator_t op, value_t dst, 
    value_t val1, value_t val2);

static const struct operator_fn_tbl_ {
  operator_t    op;
  operator_fn_t fn;
  unsigned      flags;
} operator_fn_tbl[] = {
  { operator_null,        0,                    0},
  { operator_add,         pic_op_add_sub,       0},
  { operator_sub,         pic_op_add_sub,       0},
  { operator_mul,         pic_op_mul,           0},
  { operator_div,         pic_op_div_mod,       0},
  { operator_mod,         pic_op_div_mod,       0},
  { operator_lt,          pic_op_relational,    0},
  { operator_le,          pic_op_relational,    0},
  { operator_eq,          pic_op_equality,      0},
  { operator_ne,          pic_op_equality,      0},
  { operator_ge,          pic_op_relational,    0},
  { operator_gt,          pic_op_relational,    0},
  { operator_andl,        pic_op_andl_orl,      0},
  { operator_orl,         pic_op_andl_orl,      0},
  { operator_notl,        pic_op_logical_not,   0},
  { operator_andb,        pic_op_andb_orb_xorb, 0},
  { operator_orb,         pic_op_andb_orb_xorb, 0},
  { operator_xorb,        pic_op_andb_orb_xorb, 0},
  { operator_cmpb,        pic_op_cmpb,          0},
  { operator_assign,      pic_op_assign,        0},
  { operator_neg,         pic_op_neg,           0},
  { operator_incr,        pic_op_incr_decr,     0},
  { operator_decr,        pic_op_incr_decr,     0},
  { operator_shift_left,  pic_op_shift,         0},
  { operator_shift_right, pic_op_shift,         0},
  { operator_logical,     pic_op_logical_not,   0},
  { operator_dot,         0,                    0},
  { operator_subscript,   0,                    0}
};

void pic_op(pfile_t *pf, operator_t op, value_t dst, value_t val1,
    value_t val2)
{
  variable_sz_t    init;
  size_t           ii;
  pic_last_value_t which;

  init = pic_temp_sz_in_use;
  if (operator_is_binary(op)) {
    if (!val2) {
      val2 = val1;
      val1 = dst;
    }
  } else if (operator_is_unary(op)) {
    if (!val1) {
      val1 = dst;
    }
  }
  for (ii = 0;
       (ii < COUNT(operator_fn_tbl))
       && (operator_fn_tbl[ii].op != op);
       ii++)
    ;
  if (ii == COUNT(operator_fn_tbl)) {
    assert(0);
  } else if (!operator_fn_tbl[ii].fn) {
    assert(0);
  } else {
    operator_fn_tbl[ii].fn(pf, op, dst, val1, val2);
  }
  /* if one of the last values has changed, invalid these */
  for (which = PIC_LAST_VALUE_FIRST;
       (which < PIC_LAST_VALUE_CT)
       && !value_is_same(pic_last_values[which], dst);
       which++)
    ;
  if (which < PIC_LAST_VALUE_CT) {
    pic_last_values_reset();
  }
  assert(init == pic_temp_sz_in_use);
}

